import atexit
import errno
import os
import random
import signal
import site
import socket
import sys
import time

import tornado

if tornado.version_info >= (2,1,0):
    from ._httpserver_2 import HTTPServer
else:
    from ._httpserver import HTTPServer

from zygote.util import AFUnixSender
from zygote.util import get_logger
from zygote.util import safe_kill
from zygote.util import set_nonblocking
from zygote.util import setproctitle
from zygote.util import wait_for_pids
from zygote.util import ZygoteIOLoop
import message

# Exit with this exit code when there was a failure to init the worker
# (which might be hard to represent otherwise if it, for example, occurs
# while setting up the domain socket)
INIT_FAILURE_EXIT_CODE = 4

WORKER_INIT_FAILURE_EXIT_CODE = 5

def establish_signal_handlers(logger):
    # delete atexit handlers from parent
    del atexit._exithandlers[:]

    def zygote_exit(signum, frame):
        if signum == signal.SIGINT:
            logger.info('received SIGINT, exiting')
        elif signum == signal.SIGTERM:
            logger.info('received SIGTERM, exiting')
        elif signum == signal.SIGQUIT:
            logger.info('recieved SIGQUIT (clean exit), exiting')
        else:
            logger.info('received signal %d, exiting', signum)
        sys.exit(0)
    # we explicitly ignore SIGINT and SIGTERM
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, signal.SIG_IGN)
    signal.signal(signal.SIGCHLD, signal.SIG_DFL)
    for sig in (signal.SIGQUIT,):
        signal.signal(sig, zygote_exit)

def notify(sock, msg_cls, body=''):
    """Send a message to the zygote master. Should be using AFUnixSender?"""
    sock.send(msg_cls.emit(str(body)))


class ZygoteWorker(object):
    """A Zygote is a process that manages children worker processes.

    When the zygote process is instantiated it does a few things:
     * chdirs to the absolute position pointed by a basepath symlink
     * munges sys.path to point to the new version of the code
     * imports the target module, to pre-fork load resources
     * creates read and write pipes to the parent process
    """

    RECV_SIZE = 8192

    # how many seconds to wait before sending SIGKILL to children
    WAIT_FOR_KILL_TIME = 10.0

    def __init__(self, sock, basepath, module, name, version, args, ssl_options=None, canary=False, debug=False):
        self.module = module
        self.name = name
        self.version = version
        self.basepath = basepath
        self.args = args
        self.ssl_options = ssl_options
        self.ppid = os.getppid()
        self.canary = canary
        self.children = set()
        self.debug = debug
        self.logger = get_logger('zygote.worker.zygote_process', self.debug)

        establish_signal_handlers(self.logger)

        # Set up the control socket nice and early
        try:
            self.control_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
            self.control_socket.bind('\0zygote_%d' % os.getpid())
        except Exception:
            # If we can't bind to the control socket, just give up
            self.logger.error("Could not bind to control socket, aborting early!")
            sys.exit(INIT_FAILURE_EXIT_CODE)

        try:
            self._real_init(sock, self.basepath, self.module, args)
        except Exception:
            self.logger.exception("Error performing initialization of %s", self)
            sys.exit(INIT_FAILURE_EXIT_CODE)

    def _version(self):
        """Get the actual version as specified or from the basepath."""
        return self.version or self.basepath.split('/')[-1]

    def _real_init(self, sock, basepath, module, args):
        """Actual initialization function. Broken out for error handling"""

        setproctitle('zygote name=%s version=%s' % (self.name, self._version(),))

        # Create a pipe(2) pair. This will be used so workers can detect when
        # the intermediate zygote exits -- when this happens, a read event will
        # happen on the read_pipe file descriptor, and the child can exit. We do
        # this so that if the intermediate zygote exits unexpectedly for some
        # reason, while it still has children workers running (which is an
        # abnormal situation in and of itself), we aren't left with orphaned
        # worker processes. Note that the write_pipe is normally never written
        # on, we're just using this hack to get a read event on the read pipe.
        self.read_pipe, self.write_pipe = os.pipe()

        self.io_loop = ZygoteIOLoop(log_name='zygote.worker.ioloop')

        os.chdir(basepath)

        # Add basepath to sys.path so that application will be able to
        # load what's required. We do this at zygote initialization
        # to have the correct paths at new code reload.
        site.addsitedir(os.path.realpath(basepath))

        t = __import__(module, [], [], ['initialize', 'get_application'], 0)

        self.sock = sock

        self.get_application = t.get_application

        set_nonblocking(self.control_socket)
        self.io_loop.add_handler(self.control_socket.fileno(), self.handle_control, self.io_loop.READ)

        self.notify_socket = AFUnixSender(self.io_loop)
        self.notify_socket.connect('\0zygote_%d' % self.ppid)

        signal.signal(signal.SIGCHLD, self.reap_child)

        # If there is an initialize function defined then call it.
        if hasattr(t, 'initialize'):
            self.logger.info('initializing zygote')
            t.initialize(*self.args)

        if self.canary:
            notify(self.notify_socket, message.MessageCanaryInit)
            # Initialization is successful. This is not the canary zygote anymore.
            self.canary = False

        self.logger.info('new zygote started')

    def handle_control(self, fd, events):
        assert fd == self.control_socket.fileno()
        data = self.control_socket.recv(self.RECV_SIZE)
        msg = message.Message.parse(data)
        if type(msg) is message.MessageCreateWorker:
            self.spawn_worker()
        elif type(msg) is message.MessageKillWorkers:
            self.kill_workers(msg.num_workers_to_kill)
        elif type(msg) is message.MessageShutDown:
            self.kill_all_workers()
        else:
            assert False

    def kill_workers(self, num_workers_to_kill):
        if num_workers_to_kill > len(self.children):
            self.logger.error(
                'Request to kill %d workers out of %d current workers',
                num_workers_to_kill,
                len(self.children)
            )
            return
        worker_pids = random.sample(self.children, num_workers_to_kill)
        for pid in worker_pids:
            safe_kill(pid)
        wait_for_pids(worker_pids, self.WAIT_FOR_KILL_TIME, self.logger)

    def kill_all_workers(self):
        """Kill all workers and wait (synchronously) for them
        to exit"""
        # reset the signal handler so that we don't get interrupted
        # by SIGCHLDs
        signal.signal(signal.SIGCHLD, signal.SIG_DFL)
        waiting_pids = set()

        self.logger.debug('zygote requesting kill on %d pids', len(self.children))
        for pid in self.children:
            if safe_kill(pid, signal.SIGQUIT):
                waiting_pids.add(pid)
        wait_for_pids(waiting_pids, self.WAIT_FOR_KILL_TIME, self.logger)
        self.logger.debug('zygote done killing children, terminating')
        sys.exit(0)

    def reap_child(self, signum, frame):
        assert signum == signal.SIGCHLD
        while True:
            try:
                pid, status = os.waitpid(0, os.WNOHANG)
            except OSError, e:
                if e.errno == errno.ECHILD:
                    break
                elif e.errno == errno.EINTR:
                    continue
                raise # should just be EINVAL on Linux

            if pid == 0:
                break

            status_code = os.WEXITSTATUS(status)
            self.logger.info('reaped worker %d, status %d', pid, status_code)
            if status_code == WORKER_INIT_FAILURE_EXIT_CODE:
                notify(self.notify_socket, message.MessageWorkerExitInitFail, '%d %d' % (pid, status_code))
            else:
                notify(self.notify_socket, message.MessageWorkerExit, '%d %d' % (pid, status_code))

            self.children.remove(pid)

    def loop(self):
        self.io_loop.start()

    def spawn_worker(self):
        time_created = time.time()
        pid = os.fork()
        if pid:
            self.children.add(pid)
            return
        try:
            self.logger.debug("Calling _initialize_worker")
            self._initialize_worker(time_created)
            self.logger.debug("Worker initialized")
        except Exception, e:
            self.logger.exception("Error initializing worker process: %s", e)
            sys.exit(WORKER_INIT_FAILURE_EXIT_CODE)
        self.logger.debug("Looks okay to me, smooth sailing!")

    def _initialize_worker(self, time_created):
        # We're the child. We need to close the write_pipe in order for the
        # read_pipe to get an event when the parent's write_pipe closes
        # (otherwise the kernel is too smart and thinks that it's waiting
        # for writes from *this* process' write_pipe).
        os.close(self.write_pipe)

        logger = get_logger('zygote.worker.worker_process')
        logger.debug('new worker started')

        def on_parent_exit(fd, events):
            logger.error('detected that intermediate zygote died, exiting')
            sys.exit(0)

        # create a new i/o loop
        del self.io_loop
        io_loop = ZygoteIOLoop(log_name='zygote.worker.worker_process.ioloop')
        # Install this worker's io_loop as the global io_loop; only applies in
        # this fork. Programs that uses this io_loop instance should NOT use
        # io_loop.start() because start() is invoked by the corresponding
        # zygote worker.
        if tornado.version_info >= (2,1,0):
            io_loop.install()
        else:
            tornado.ioloop.IOLoop._instance = io_loop

        # add the read pipe
        io_loop.add_handler(self.read_pipe, on_parent_exit, io_loop.READ)

        sock = AFUnixSender(io_loop, logger=logger)
        sock.connect('\0zygote_%d' % self.ppid)

        establish_signal_handlers(logger)
        def on_headers(line, remote_ip, headers):
            logger.debug('sending MessageHTTPBegin')
            notify(sock, message.MessageHTTPBegin, "%s %s" % (remote_ip, line))
        def on_close(disconnected=False):
            logger.debug('sending MessageHTTPEnd')
            notify(sock, message.MessageHTTPEnd)

        notify(sock, message.MessageWorkerStart, '%d %d' % (int(time_created * 1e6), os.getppid()))
        setproctitle('zygote-worker name=%s version=%s' % (self.name, self._version(),))
        try:
            # io_loop is passed into get_application for program to add handler
            # or schedule task on the main io_loop.  Program that uses this
            # io_loop instance should NOT use io_loop.start() because start()
            # is invoked by the corresponding zygote worker.
            kwargs = {'io_loop': io_loop}
            logger.debug("Invoking get_application")
            app = self.get_application(*self.args, **kwargs)
        except Exception:
            logger.error("Unable to get application")
            raise
        # TODO: make keep-alive servers work
        logger.debug("Creating HTTPServer")
        http_server = HTTPServer(app,
                io_loop=io_loop,
                no_keep_alive=True,
                close_callback=on_close,
                headers_callback=on_headers,
                ssl_options=self.ssl_options
        )
        if tornado.version_info >= (2,1,0):
            http_server.add_socket(self.sock)
        else:
            http_server._socket = self.sock
            io_loop.add_handler(self.sock.fileno(), http_server._handle_events, io_loop.READ)
        logger.debug("Started ioloop...")
        io_loop.start()



# -*- coding: utf-8 -*-
import pkg_resources


# We should define the package version external to
# it. http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
version = pkg_resources.require('zygote')[0].version



import atexit
import datetime
import errno
import fcntl
import logging
import os
import signal
import socket
import struct
import sys
import time

import tornado.ioloop

from zygote import accounting
from zygote import handlers
from zygote import message
from zygote.util import close_fds
from zygote.util import safe_kill
from zygote.util import setproctitle
from zygote.util import wait_for_pids
from zygote.util import ZygoteIOLoop
from zygote.util import get_logger
from zygote.util import NullHandler
from zygote.util import LocklessHandler
from zygote.worker import INIT_FAILURE_EXIT_CODE
from zygote.worker import ZygoteWorker


try:
    import ssl # Python 2.6+
except ImportError:
    ssl = None


class ZygoteMaster(object):

    instantiated = False

    RECV_SIZE = 8192

    # number of seconds to wait between polls
    POLL_INTERVAL = 1.0

    # how many seconds to wait before sending SIGKILL to children
    WAIT_FOR_KILL_TIME = 10.0

    def __init__(
        self,
        sock,
        basepath,
        module,
        name,
        version,
        num_workers,
        control_port,
        control_socket_path,
        application_args=None,
        max_requests=None,
        zygote_base=None,
        ssl_options=None,
        debug=False
    ):
        self.logger = get_logger('zygote.master', debug)
        if self.__class__.instantiated:
            self.logger.error('cannot instantiate zygote master more than once')
            sys.exit(1)
        self.__class__.instantiated = True

        self.sock = sock
        self.basepath = basepath
        self.module = module
        self.name = name
        self.version = version
        self.num_workers = num_workers
        self.control_port = control_port
        self.control_socket_path = control_socket_path
        self.application_args = application_args or []
        self.max_requests = max_requests
        self.zygote_base = zygote_base
        self.ssl_options = ssl_options
        self.debug = debug

        self.stopped = False
        self.started_transition = None
        self.prev_zygote = None
        self.current_zygote = None
        self.time_created = datetime.datetime.now()
        self.io_loop = ZygoteIOLoop(log_name='zygote.master.ioloop')
        self.zygote_collection = accounting.ZygoteCollection()

        self.setup_master_socket()
        self.setup_control_socket()

        signal.signal(signal.SIGCHLD, self.reap_child)
        signal.signal(signal.SIGHUP, self.update_revision)
        for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT):
            signal.signal(sig, self.stop)

        self.open_fds, self.status_http_server = handlers.get_httpserver(
                self.io_loop,
                self.control_port,
                self,
                zygote_base=self.zygote_base,
                ssl_options=self.ssl_options,
        )

    def setup_master_socket(self):
        """Create an abstract unix domain socket for master. This
        socket will be used to receive messages from zygotes and their
        children.
        """
        self.logger.debug("Binding to master domain socket")
        self.master_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
        self.master_socket.bind('\0zygote_%d' % os.getpid())
        self.io_loop.add_handler(self.master_socket.fileno(), self.handle_protocol_msg, self.io_loop.READ)

    def setup_control_socket(self):
        try:
            socket_path = self.control_socket_path
            if os.path.exists(socket_path):
                # NOTE: Starting the same application twice we won't get
                # here since main() won't be able to bind. We can add a
                # (ex|nb) file lock if needed.
                self.logger.error("Control socket exitsts %s. Probably from a previous run. Removing...", socket_path)
                self.cleanup_control_socket()
            self.logger.debug("Binding to control socket %s", socket_path)
            self.control_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
            self.control_socket.bind(socket_path)
            self.io_loop.add_handler(self.control_socket.fileno(), self.handle_control_msg, self.io_loop.READ)
        except Exception, e:
            # This is treated as a fatal error as only way to recover
            # from this is to restart zygote master and it's not what
            # we want.
            self.logger.error("Can not bind to control socket: %s", e)
            self.logger.error("Control socket is needed to make configuration changes on the running zygote master.")
            sys.exit(1)

    def cleanup_control_socket(self):
        if os.path.exists(self.control_socket_path):
            self.logger.debug("Removing control socket at %s" % self.control_socket_path)
            os.unlink(self.control_socket_path)

    def handle_control_msg(self, fd, events):
        assert fd == self.control_socket.fileno()
        data = self.control_socket.recv(self.RECV_SIZE)
        msg = message.ControlMessage.parse(data)
        msg_type = type(msg)

        # NOTE: We can possibly use SO_PEERCRED on control socket to
        # get more information about the client.
        self.logger.info('received message of type %s', msg_type.__name__,)

        if msg_type is message.ControlMessageScaleWorkers:
            self.scale_workers(msg.num_workers)

    def handle_protocol_msg(self, fd, events):
        """Callback for messages received on the master_socket"""
        assert fd == self.master_socket.fileno()
        data = self.master_socket.recv(self.RECV_SIZE)
        msg = message.Message.parse(data)
        msg_type = type(msg)
        self.logger.debug('received message of type %s from pid %d', msg_type.__name__, msg.pid)

        if msg_type is message.MessageCanaryInit:
            self.logger.info("Canary zygote initialized. Transitioning idle workers.")
            # This is not the canary zygote anymore
            self.current_zygote.canary = False
            # We can also release the handle on the previous
            # zygote. It is already in the zygote_collection for
            # accounting purposses, but we won't need to keep track of
            # it anymore.
            self.prev_zygote = None
            # Canary initialization was successful, we can now transition workers
            self.io_loop.add_callback(self.transition_idle_workers)
        elif msg_type is message.MessageWorkerStart:
            # a new worker was spawned by one of our zygotes; add it to
            # zygote_collection, and note the time created and the zygote parent
            zygote = self.zygote_collection[msg.worker_ppid]
            if zygote:
                zygote.add_worker(msg.pid, msg.time_created)
        elif msg_type is message.MessageWorkerExitInitFail:
            if not self.current_zygote.canary:
                self.logger.error("A worker initialization failed, giving up")
                self.stop()
                return
        elif msg_type is message.MessageWorkerExit:
            # a worker exited. tell the current/active zygote to spawn a new
            # child. if this was the last child of a different (non-current)
            # zygote, kill that zygote
            zygote = self.zygote_collection[msg.pid]
            if not zygote:
                return

            zygote.remove_worker(msg.child_pid)
            if zygote.shutting_down:
                self.logger.debug('Removed a worker from shutting down zygote %d, %d left', msg.pid, len(zygote.workers()))
                return
            else:
                self.logger.debug('Removed a worker from zygote %d, %d left', msg.pid, len(zygote.workers()))

            if not self.stopped:
                if zygote in (self.current_zygote, self.prev_zygote):
                    if self.num_workers > zygote.worker_count:
                        # Only start a new if we're below quota. This
                        # is how we scale down the number of workers.
                        zygote.request_spawn()
                else:
                    # Not a zygote that we care about. Request shutdown.
                    zygote.request_shut_down()
        elif msg_type is message.MessageHTTPBegin:
            # a worker started servicing an HTTP request
            worker = self.zygote_collection.get_worker(msg.pid)
            if worker:
                worker.start_request(msg.remote_ip, msg.http_line)
        elif msg_type is message.MessageHTTPEnd:
            # a worker finished servicing an HTTP request
            worker = self.zygote_collection.get_worker(msg.pid)
            if worker:
                worker.end_request()
                if self.max_requests is not None and worker.request_count >= self.max_requests:
                    self.logger.info('Worker %d reached max_requests %d, killing it', worker.pid, self.max_requests)
                    safe_kill(worker.pid, signal.SIGQUIT)
        else:
            self.logger.warning('master got unexpected message of type %s', msg_type)


    def scale_workers(self, num_workers):
        prev_num_workers = self.num_workers
        diff_num_workers = num_workers - self.num_workers
        self.num_workers = num_workers
        if not diff_num_workers:
            return
        elif diff_num_workers > 0:
            self.logger.info('Increasing number of workers from %d to %d.', prev_num_workers, num_workers)
            for _ in range(diff_num_workers):
                self.current_zygote.request_spawn()
        else:
            self.logger.info('Reducing number of workers from %d to %d.', prev_num_workers, num_workers)
            self.current_zygote.request_kill_workers(-diff_num_workers)

    def reap_child(self, signum, frame):
        """Signal handler for SIGCHLD. Reaps children and updates
        self.zygote_collection.
        """
        assert signum == signal.SIGCHLD
        while True:
            try:
                # The Zygotes are in their own process group, so need to
                # call waitpid() with -1 instead of 0. See waitpid(2).
                pid, status = os.waitpid(-1, os.WNOHANG)
            except OSError, e:
                if e.errno == errno.ECHILD:
                    break
                elif e.errno == errno.EINTR:
                    continue
                raise
            if pid == 0:
                break

            status_code = os.WEXITSTATUS(status)
            self.logger.info('zygote %d exited with status %d', pid, status_code)

            # the zygote died. if the zygote was not the current zygote it's OK;
            # otherwise, we need to start a new one
            try:
                self.zygote_collection.remove_zygote(pid)
            except KeyError:
                pass

            if status_code == INIT_FAILURE_EXIT_CODE:
                if pid == self.current_zygote.pid and self.current_zygote.canary:
                    if self.prev_zygote:
                        self.curent_zygote = self.prev_zygote
                    self.logger.error("Could not initialize canary worker. Giving up trying to respawn")
                else:
                    self.logger.error("Could not initialize zygote worker, giving up")
                    self.really_stop()
                return

            if not self.stopped:
                active_zygote = self.current_zygote

                if pid == self.current_zygote.pid:
                    self.current_zygote = self.create_zygote()
                    active_zygote = self.current_zygote
                elif self.prev_zygote and pid == self.prev_zygote.pid:
                    self.prev_zygote = self.create_zygote()
                    active_zygote = self.prev_zygote

                # we may need to create new workers for the active zygote... this
                # is a bit racy, although that seems to be pretty unlikely in
                # practice
                workers_needed = self.num_workers - self.zygote_collection.worker_count()
                for x in xrange(workers_needed):
                    active_zygote.request_spawn()

            elif len(self.zygote_collection.zygote_map.values()) == 0:
                self.really_stop()

    def stop(self, signum=None, frame=None):
        """
        Stop the zygote master. Steps:
          * Ask all zygotes to kill and wait on their children
          * Wait for zygotes to exit
          * Kill anything left over if necessary
        """
        if self.stopped:
            return
        # kill all of the workers
        self.logger.info('stopping all zygotes and workers')
        pids = set()
        for zygote in self.zygote_collection:
            pids.add(zygote.pid)
            self.logger.debug('requesting shutdown on %d', zygote.pid)
            zygote.request_shut_down()

        self.logger.debug('setting self.stopped')
        self.stopped = True

        self.logger.debug('master is stopping. will not try to update anymore.')
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

        self.logger.debug('stopping io_loop.')
        if getattr(self, 'io_loop', None) is not None:
            self.io_loop.stop()

        self.logger.info('waiting for workers to exit before stoping master.')
        wait_for_pids(pids, self.WAIT_FOR_KILL_TIME, self.logger, kill_pgroup=True)
        self.logger.info('all zygotes exited; good night')

        self.really_stop(0)

    def really_stop(self, status=0):
        self.cleanup_control_socket()
        sys.exit(status)

    def transition_idle_workers(self):
        """Transition idle HTTP workers from old zygotes to the current
        zygote.
        """
        if not self.started_transition:
            self.started_transition = time.time()
        if (time.time() - self.started_transition) > self.WAIT_FOR_KILL_TIME:
            self.logger.debug("sending SIGKILL for transition because it was Too Damn Slow")
            sig = signal.SIGKILL
        else:
            sig = signal.SIGQUIT

        other_zygotes = self.zygote_collection.other_zygotes(self.current_zygote)
        if self.current_zygote.canary and self.prev_zygote:
            if self.prev_zygote in other_zygotes:
                other_zygotes.remove(self.prev_zygote)

        kill_count = 0
        other_zygote_count = len(other_zygotes)
        for zygote in other_zygotes:
            for worker in zygote.idle_workers():
                self.logger.debug("killing worker %d with signal %d", worker.pid, sig)
                if safe_kill(worker.pid, sig):
                    kill_count += 1
        self.logger.info('Attempted to transition %d workers from %d zygotes', kill_count, other_zygote_count)

        if other_zygote_count:
            # The list of other zygotes was at least one, so we should
            # reschedule another call to transition_idle_workers. When a zygote
            # runs out of worker children, the handle_protocol_msg function will
            # notice this fact when it receives the final MessageWorkerExit, and
            # at that time it will kill the worker, which is how this timeout
            # loop gets ended.
            self.io_loop.add_timeout(time.time() + self.POLL_INTERVAL, self.transition_idle_workers)
        else:
            self.started_transition = None

        # Cleanup empty zygotes for the next iteration of the transition.
        for zygote in other_zygotes:
            if zygote.worker_count == 0:
                self.kill_empty_zygote(zygote, sig)

    def kill_empty_zygote(self, zygote, sig=signal.SIGQUIT):
        """Send zygote SIGQUIT if it has zero workers. """
        # The only valid time to kill a zygote is if it doesn't have
        # any workers left.
        if zygote.worker_count == 0:
            self.logger.info("killing zygote with pid %d" % zygote.pid)
            safe_kill(zygote.pid, sig)

    def update_revision(self, signum=None, frame=None):
        """The SIGHUP handler, calls create_zygote and possibly initiates the
        transition of idle workers.

        This preserves the current zygote and initializes a "canary"
        zygote as the current one.
        """
        self.prev_zygote = self.current_zygote
        self.current_zygote = self.create_zygote(canary=True)

    def create_zygote(self, canary=False):
        """"Create a new zygote"""
        # read the basepath symlink
        realbase = os.path.realpath(self.basepath)

        pid = os.fork()
        if pid:
            self.logger.info('started zygote %d pointed at base %r', pid, realbase)
            z = self.zygote_collection.add_zygote(pid, realbase, self.io_loop, canary=canary)
            if not canary:
                self.io_loop.add_callback(self.transition_idle_workers)
            return z
        else:
            # Try to clean up some of the file descriptors and whatnot that
            # exist in the parent before continuing. Strictly speaking, this
            # isn't necessary, but it seems good to remove these resources
            # if they're not needed in the child.
            del self.io_loop
            close_fds(self.sock.fileno())
            signal.signal(signal.SIGHUP, signal.SIG_DFL)

            # Make the zygote a process group leader
            os.setpgid(os.getpid(), os.getpid())
            # create the zygote
            z = ZygoteWorker(
                    sock=self.sock,
                    basepath=realbase,
                    module=self.module,
                    name=self.name,
                    version=self.version,
                    args=self.application_args,
                    ssl_options=self.ssl_options,
                    canary=canary,
                    debug=self.debug
            )
            z.loop()

    def start(self):
        self.current_zygote = self.create_zygote()
        for x in xrange(self.num_workers):
            self.current_zygote.request_spawn()
        self.io_loop.start()

def main(opts, extra_args):
    setproctitle('zygote master %s' % (opts.name or opts.module,))
    zygote_logger = get_logger('zygote', opts.debug)

    if not logging.root.handlers:
        # XXX: WARNING
        #
        # We're disabling the root logger. Tornado's RequestHandler ONLY
        # supports logging uncaught errors to the root logger. This will end
        # poorly for you!
        #
        # We should probably provide a RequestHandler subclass that has
        # _handle_request_exception overridden to do something useful.
        # That might be hard to do without adding a tight version dependency
        # on tornado.
        logging.root.addHandler(NullHandler())

    if opts.debug:
        logging.root.setLevel(logging.DEBUG)
    else:
        logging.root.setLevel(logging.INFO)

    zygote_logger.info('main started')

    # Create the TCP listen socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    flags = fcntl.fcntl(sock.fileno(), fcntl.F_GETFD)
    flags |= fcntl.FD_CLOEXEC
    fcntl.fcntl(sock.fileno(), fcntl.F_SETFD, flags)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setblocking(0)
    sock.bind((opts.interface, opts.port))
    sock.listen(128)

    ssl_options=None
    if opts.cert:
        ssl_options = dict(
                certfile=opts.cert,
                keyfile=opts.key,
                ca_certs=opts.cacerts,
                cert_reqs=ssl.CERT_OPTIONAL if opts.cacerts else ssl.CERT_NONE,
        )
        zygote_logger.info('using SSL with %s', ssl_options)

        sock = ssl.wrap_socket(sock,
                server_side=True,
                do_handshake_on_connect=False,
                **ssl_options
        )

    master = ZygoteMaster(
        sock,
        basepath=opts.basepath,
        module=opts.module,
        name=opts.name or opts.module,
        version=opts.version,
        num_workers=opts.num_workers,
        control_port=opts.control_port,
        control_socket_path=opts.control_socket_path,
        application_args=extra_args,
        max_requests=opts.max_requests,
        zygote_base=opts.zygote_base,
        ssl_options=ssl_options,
        debug=opts.debug
    )
    atexit.register(master.stop)
    master.start()



#!/usr/bin/env python
#
# Copyright 2011, Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import optparse
import os
import sys

import zygote.master
import zygote.util

def main():
    usage = 'usage: %prog -b <basepath> -m <module> -p <port> [module_args...]'
    parser = optparse.OptionParser(usage=usage)
    parser.add_option('-b', '--basepath', default=os.environ.get('BASEPATH', ''), help='The basepath to use')
    parser.add_option('--control-port', type='int', default=5100, help='The control port to listen on')
    parser.add_option('-d', '--debug', default=False, action='store_true', help='Enable debugging')
    parser.add_option('-n', '--name', default=None, help='The name of the application to set in proctitle, otherwise use the app module.')
    parser.add_option('--version', default=None, help='The version of the application to set in proctitle.')
    parser.add_option('-m', '--module', default=None, help='The name of the module holding get_application()')
    parser.add_option('-p', '--port', type='int', default=0, help='The port to bind on')
    parser.add_option('-i', '--interface', default='', help='The interface to bind on')
    parser.add_option('--num-workers', type='int', default=8, help='How many workers to run')
    parser.add_option('--max-requests', type='int', default=None, help='The maximum number of requests a child can run')
    parser.add_option('--zygote-base', default=None, help='The base path to the zygote')
    parser.add_option('--cert', default=None, help='Certificate to use for HTTPS traffic')
    parser.add_option('--key', default=None, help='Private key for HTTPS traffic')
    parser.add_option('--cacerts', default=None, help='File containing a list of root certificates')
    parser.add_option(
        '--control-socket',
        dest='control_socket_path',
        default=os.path.join(zygote.util.get_rundir(), "zygote_master.sock"),
        help='The socket to control zygote master at run time'
    )

    opts, args = parser.parse_args()

    if not opts.basepath:
        parser.error('The `basepath` cannot be empty; specify one with -b or exporting BASEPATH')
        sys.exit(1)
    if not opts.port:
        parser.error('No port was specified')
        sys.exit(1)
    if not opts.module:
        parser.error('You must specify a module argument using -m')
        sys.exit(1)

    zygote.master.main(opts, args)

if __name__ == '__main__':
    main()



import os

class Message(object):

    CANARY_INIT           = 'I'

    CREATE_WORKER         = 'C'
    KILL_WORKERS          = 'L'
    SHUT_DOWN             = 'K'

    WORKER_START          = 'S'
    WORKER_EXIT           = 'X'
    WORKER_EXIT_INIT_FAIL = 'Y'

    HTTP_BEGIN            = 'B'
    HTTP_END              = 'E'

    @classmethod
    def emit(cls, body):
        return '%d %s %s' % (os.getpid(), cls.msg_type, body)

    @classmethod
    def parse(cls, msg):
        pid, type, body = msg.split(' ', 2)
        pid = int(pid)
        if type == cls.CREATE_WORKER:
            return MessageCreateWorker(pid, body)
        elif type == cls.KILL_WORKERS:
            return MessageKillWorkers(pid, body)
        elif type == cls.CANARY_INIT:
            return MessageCanaryInit(pid, body)
        elif type == cls.WORKER_START:
            return MessageWorkerStart(pid, body)
        elif type == cls.WORKER_EXIT:
            return MessageWorkerExit(pid, body)
        elif type == cls.HTTP_BEGIN:
            return MessageHTTPBegin(pid, body)
        elif type == cls.HTTP_END:
            return MessageHTTPEnd(pid, body)
        elif type == cls.WORKER_EXIT_INIT_FAIL:
            return MessageWorkerExitInitFail(pid, body)
        elif type == cls.SHUT_DOWN:
            return MessageShutDown(pid, body)
        else:
            assert False

    def __init__(self, pid):
        self.pid = int(pid)

class MessageCanaryInit(Message):

    msg_type = Message.CANARY_INIT

    def __init__(self, pid, body):
        assert body == ''
        super(MessageCanaryInit, self).__init__(pid)

class MessageCreateWorker(Message):

    msg_type = Message.CREATE_WORKER

    def __init__(self, pid, body):
        assert body == ''
        super(MessageCreateWorker, self).__init__(pid)

class MessageKillWorkers(Message):

    msg_type = Message.KILL_WORKERS

    def __init__(self, pid, body):
        super(MessageKillWorkers, self).__init__(pid)
        self.num_workers_to_kill = int(body)

class MessageWorkerStart(Message):

    msg_type = Message.WORKER_START

    def __init__(self, pid, body):
        super(MessageWorkerStart, self).__init__(pid)
        created, ppid = body.split(' ')
        self.time_created = int(created)
        self.worker_ppid = int(ppid)

class MessageWorkerExit(Message):

    msg_type = Message.WORKER_EXIT

    def __init__(self, pid, body):
        super(MessageWorkerExit, self).__init__(pid)
        child_pid, status = body.split()
        self.payload = body
        self.child_pid = int(child_pid)
        self.status = int(status)

class MessageWorkerExitInitFail(Message):

    msg_type = Message.WORKER_EXIT_INIT_FAIL

    def __init__(self, pid, body):
        super(MessageWorkerExitInitFail, self).__init__(pid)
        child_pid, status = body.split()
        self.payload = body
        self.child_pid = int(child_pid)
        self.status = int(status)

class MessageHTTPBegin(Message):

    msg_type = Message.HTTP_BEGIN

    def __init__(self, pid, body):
        super(MessageHTTPBegin, self).__init__(pid)
        self.remote_ip, self.http_line = body.split(' ', 1)

class MessageHTTPEnd(Message):

    msg_type = Message.HTTP_END

    def __init__(self, pid, body):
        super(MessageHTTPEnd, self).__init__(pid)
        assert body == '' # just ignore the body, it should be empty

class MessageShutDown(Message):

    msg_type = Message.SHUT_DOWN

    def __init__(self, pid, body):
        super(MessageShutDown, self).__init__(pid)
        assert body == '' # just ignore the body, it should be empty


class ControlMessage(object):
    """Control Message type used by zygote master"""

    UNKNOWN = 'I'
    SCALE_WORKERS = 'S'

    @classmethod
    def emit(cls, body):
        return '%s %s' % (cls.msg_type, body)

    @classmethod
    def parse(cls, msg):
        type, body = msg.split(' ', 1)
        if type == cls.SCALE_WORKERS:
            return ControlMessageScaleWorkers(body)
        else:
            return ControlMessageUnknown(body)

class ControlMessageUnknown(ControlMessage):

    msg_type = ControlMessage.UNKNOWN

    def __init__(self, body):
        super(ControlMessageUnknown, self).__init__()

class ControlMessageScaleWorkers(ControlMessage):

    msg_type = ControlMessage.SCALE_WORKERS

    def __init__(self, body):
        super(ControlMessageScaleWorkers, self).__init__()
        self.num_workers = int(body)



import datetime
import os
import signal
import time

import zygote.util
from zygote import message
from zygote.util import meminfo_fmt

log = zygote.util.get_logger('zygote.accounting')

def format_millis(v):
    if v is not None:
        return '%1.1f' % (v * 1000.0)

class Worker(object):

    def __init__(self, pid, time_created=None):
        if time_created:
            self.time_created = datetime.datetime.fromtimestamp(time_created / 1e6)
        else:
            self.time_created = datetime.datetime.now()
        self.pid = pid
        self.vsz = ''
        self.rss = ''
        self.shr = ''
        self.remote_ip = None
        self.request_count = 0
        self.request_started = None
        self.http = None

    def update_meminfo(self):
        for k, v in meminfo_fmt(self.pid).iteritems():
            setattr(self, k, v)

    def __eq__(self, other_pid):
        return self.pid == other_pid

    def start_request(self, remote_ip, http):
        self.remote_ip = remote_ip
        self.request_started = time.time()
        self.request_count += 1
        self.http = http

    def end_request(self):
        self.remote_ip = None
        self.http = None
        self.request_started = None

    def to_dict(self):
        d = {'pid': self.pid,
             'vsz': self.vsz,
             'rss': self.rss,
             'shr': self.shr,
             'time_created': self.time_created,
             'remote_ip': self.remote_ip,
             'request_count': self.request_count,
             'http': self.http}
        if self.request_started is None:
            d['elapsed'] = None
            d['elapsed_formatted'] = None
        else:
            now = time.time()
            d['elapsed'] = now - self.request_started
            d['elapsed_formatted'] = format_millis(d['elapsed'])
        return d

    def request_exit(self):
        """Instruct this worker to exit"""
        os.kill(self.pid, signal.SIGTERM)

class Zygote(object):
    """Stub representing the zygote from the master side of the fork. Is not
    actually the zygote, but sends some commands over the unix domain socket
    to the zygote.

    TODO: Move parsing of messages *from* the unix domain socket into this object,
    and use a regular callback system
    """
    __generation = 0

    def __init__(self, pid, basepath, io_loop, canary=False):
        """Initialize using real Zygote's pid, basepath and master's
        io_loop.

        Master also marks the zygote as 'canary' at initialization if
        it's an update to a newer revision of the source. Only if the
        canary is live, master continues on transition workers.
        """
        self.basepath = basepath
        self.pid = pid
        self.worker_map = {}
        self.time_created = datetime.datetime.now()
        self.vsz = ''
        self.rss = ''
        self.shr = ''
        self.connected = False
        self.send_queue = []
        self.write_queue_active = False
        self.canary = canary

        # wait until the control_socket can be connected, since it might take a
        # moment before the forked child creates their socket. a better way to
        # do this would be to have the parent create the control_socket and then
        # the child inherits it through forking
        self.control_socket = zygote.util.AFUnixSender(io_loop)
        self.control_socket.connect('\0zygote_%d' % self.pid)

        self.shutting_down = False

        self.generation = self.__class__.__generation
        self.__class__.__generation += 1

    def update_meminfo(self):
        for k, v in meminfo_fmt(self.pid).iteritems():
            setattr(self, k, v)
        for worker in self.worker_map.itervalues():
            worker.update_meminfo()

    def workers(self):
        return self.worker_map.values()

    def add_worker(self, pid, time_created=None):
        worker = Worker(pid, time_created)
        self.worker_map[pid] = worker

    def remove_worker(self, pid):
        try:
            del self.worker_map[pid]
        except KeyError:
            log.warning("Tried to delete unknown worker %d (did worker initialization fail?)", pid)

    def begin_http(self, pid, http):
        self.worker_map[pid].http = http

    def end_http(self, pid, http):
        self.worker_map[pid].http = None

    def idle_workers(self):
        return [w for w in self.worker_map.itervalues() if w.http is None]

    def get_worker(self, pid):
        return self.worker_map.get(pid)

    def request_spawn(self):
        """Instruct this zygote to spawn a new worker"""
        log.debug('requesting spawn on Zygote %d', self.generation)
        self.control_socket.send(message.MessageCreateWorker.emit(''))

    def request_kill_workers(self, num_workers_to_kill):
        """Instruct this zygote to kill an idle worker"""
        self.control_socket.send(message.MessageKillWorkers.emit('%d' % num_workers_to_kill))

    def request_shut_down(self):
        """Instruct this zygote to shut down all workers"""
        self.control_socket.send(message.MessageShutDown.emit(""))
        self.shutting_down = True

    @property
    def worker_count(self):
        return len(self.worker_map)

    def to_dict(self):
        return {
            'basepath': self.basepath,
            'pid': self.pid,
            'workers': sorted(self.worker_map.values(), key=lambda x: x.time_created),
            'vsz': self.vsz,
            'rss': self.rss,
            'shr': self.shr,
            'time_created': self.time_created,
            'generation': self.generation,
            }

class ZygoteCollection(object):

    def __init__(self):
        self.zygote_map = {}

    def add_zygote(self, pid, basepath, io_loop, canary=False):
        z = Zygote(pid, basepath, io_loop, canary=canary)
        self.zygote_map[pid] = z
        return z

    def update_meminfo(self):
        for z in self.zygote_map.values():
            z.update_meminfo()

    def remove_zygote(self, pid):
        del self.zygote_map[pid]

    def get_worker(self, pid):
        for zygote in self.zygote_map.itervalues():
            w = zygote.get_worker(pid)
            if w:
                return w
        return None

    def basepath_to_zygote(self, basepath):
        for zygote in self.zygote_map.itervalues():
            if zygote.basepath == basepath:
                return zygote
        return None

    def __getitem__(self, pid):
        # ZygoteMaster requests a zygote using it's pid when it
        # recieves a message from the zygote. In certain cases,
        # ZygoteMaster can request a zygote that is already
        # removed. This can happen if zygote dies and master handles
        # the signal accordingly before reading the message from
        # socket.
        return self.zygote_map.get(pid, None)

    def __iter__(self):
        return self.zygote_map.itervalues()

    def other_zygotes(self, target):
        return [z for z in self.zygote_map.itervalues() if z != target]

    def to_dict(self):
        return {'zygotes': sorted(self.zygote_map.values(), key=lambda x: x.time_created)}

    def pids(self):
        return self.zygote_map.keys()

    def worker_count(self):
        """Return the total number of workers"""
        return sum(len(z.workers()) for z in self.zygote_map.itervalues())



# -*- coding: utf-8 -*-
from __future__ import with_statement

import errno
import fcntl
import functools
import logging
import os
import resource
import signal
import socket
import sys
import time

import tornado.ioloop

try:
    import setproctitle as _setproctitle
    has_proc_title = True
except ImportError:
    has_proc_title = False

log = logging.getLogger('zygote.util')

def setproctitle(name):
    if has_proc_title:
        _setproctitle.setproctitle(name)

def is_eintr(exc):
    """Returns True if an exception is an EINTR, False otherwise."""
    if hasattr(exc, 'errno'):
        return exc.errno == errno.EINTR
    elif getattr(exc, 'args', None) and hasattr(exc, 'message'):
        return exc.args[0] == errno.EINTR
    return False

def set_nonblocking(fd):
    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

def get_rundir():
    return os.getenv('RUNDIR') if os.getenv('RUNDIR') else os.getcwd()

def get_meminfo(pid=None):
    """Get the memory statistics for the current process. Values are returned
    as kilobytes. The meanings of the fields are:
      virt -- virtual size
      res -- RSS size
      shr -- shared memory
      trs -- kilobytes from 'code' pages
      drs -- kilobytes from data/stack pages
      lrs -- kilobytes from library pages
      dt -- kilobytes from diry pages
    """
    try:
        with open('/proc/%d/statm' % (pid or os.getpid())) as raw_file:
            data = raw_file.read().rstrip('\n')
    except IOError:
        return dict()

    fields = ['virt', 'res', 'shr', 'trs', 'drs', 'lrs', 'dt']
    pagesize = resource.getpagesize()
    return dict((k, int(v) * pagesize >> 10) for k, v in zip(fields, data.split()))

def meminfo_fmt(pid=None):
    d = get_meminfo(pid)
    return {
        'rss': '%1.2f' % (d['res'] / 1024.0),
        'vsz': '%1.2f' % (d['virt'] / 1024.0),
        'shr': '%1.2f' % (d['shr'] / 1024.0)
        }

def retry_eintr(func, max_retries=5):
    """Retry a function on EINTR"""
    for x in xrange(max_retries):
        try:
            return func()
        except Exception, e:
            if not is_eintr(e) or x == max_retries - 1:
                raise

def close_fds(*exclude):
    """Try to close open file descriptors. This will probably only work on
    Linux, since it uses /proc/PID/fd to get information on what file
    descriptors to close.

    An alternative for non-Linux systems would be to just try to close random
    file descriptors (say, the first 16k), but it doesn't seem like it's really
    worth the trouble (and doing this is potentially slow).
    """
    return # XXX: fixme
    if not os.path.exists('/proc/self/fd'):
        log.warn('no /proc fd information running, not closing fds')
        return
    excl = list(exclude) + [sys.stdin.fileno(), sys.stdout.fileno(), sys.stderr.fileno()]
    for fd_name in os.listdir('/proc/self/fd'):
        fd = int(fd_name)
        if fd not in excl:
            try:
                retry_eintr(lambda: os.close(fd))
            except OSError, e:
                if e.errno == errno.EBADF:
                    # for some reason the fd was bad. nothing we can do about
                    # that
                    pass
                else:
                    raise

def is_pid_alive(pid):
    """Sends null signal to a process to check if it's alive"""
    try:
        os.kill(pid, 0)
        return True
    except OSError, e:
        # Access denied, but process is alive
        return e.errno == errno.EPERM
    except:
        return False

def safe_kill(pid, sig=signal.SIGUSR1, process_group=False):
    try:
        log.debug('killing %d', pid)
        if process_group:
            os.killpg(pid, sig)
        else:
            os.kill(pid, sig)
    except OSError, e:
        # Process may have died before we send the signal
        if not is_pid_alive(pid):
            return True
        log.warning('failed to safe_kill pid %d because of %r' % (pid, e))
        return False
    return True

def wait_for_pids(pids, timeout, log, kill_pgroup=False):
    """Wait for the given Set of pids to die. If they
    haven't died after timeout seconds, send them all SIGKILL.
    
    If kill_pgroup is true, the kill will be sent to
    the process group instead of to the process itself.
    """
    start = time.time()
    elapsed = 0
    while elapsed < timeout:
        if not pids:
            break
        pid, _, _ = os.wait3(os.WNOHANG)
        if pid == 0:
            time.sleep(0.5)
        else:
            pids.remove(pid)
        elapsed = time.time() - start
    else:
        # will only get here if we ran out of time
        log.warning("PIDs [%s] didn't quit after %f seconds, sending SIGKILL", ",".join(str(p) for p in pids), timeout)
        for pid in pids:
            safe_kill(pid, signal.SIGKILL, kill_pgroup)

class AFUnixSender(object):
    """Sender abstraction for an AF_UNIX socket (using the SOCK_DGRAM
    protocol). This handles connecting in a non-blocking fashion, and sending
    messages asynchronously. Messages that are scheduled to be sent before the
    socket is connected will be sent once the socket finishes connecting.
    """

    CONNECT_FREQUENCY = 0.1

    def __init__(self, io_loop, sock=None, logger=None):
        self.io_loop = io_loop
        if sock is None:
            self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)
        else:
            self.socket = sock
        set_nonblocking(self.socket)
        self.log = logger or log

        self.connected = False # is the socket connected>
        self.send_queue = []   # queue of messages to send
        self.sending = False   # are there queued messages?

    def connect(self, target):
        try:
            self.socket.connect(target)
        except socket.error, e:
            if e.errno == errno.EINPROGRESS:
                # usual case -- the nonblocking connect causes EINPROGRESS. When
                # the socket is writeable, then the connect has finished, and we
                # call _finish_connecting
                self.io_loop.add_handler(self.socket.fileno(), self._finish_connecting, self.io_loop.WRITE)
            elif e.errno == errno.ECONNREFUSED:
                # the connection was refused. Retry the connection in
                # CONNECT_FREQUENCY seconds
                self.io_loop.add_timeout(time.time() + self.CONNECT_FREQUENCY, lambda: self.connect(target))
            else:
                raise
        else:
            # we were able to connect immediately
            self._finish_connecting()

    def _finish_connecting(self, fd=None, events=None):
        if fd is not None:
            assert fd == self.socket.fileno()
            self.io_loop.remove_handler(fd)
            error = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
            if error != 0:
                self.log.error('got socket connect error %r' % (error,))
                raise IOError('errno %d' % (error,))

        self.connected = True
        if self.send_queue:
            self._sendall()

    def _sendall(self):
        if not self.connected:
            return
        if not self.send_queue:
            self.log.error('got _sendall with no send_queue')
            return
        if self.sending:
            # could happen if we schedule multiple messages to be sent in a row
            self.log.debug('already in send loop, be patient')
            return

        def maybe_send_queue(fd, _):
            """Try to send the message queue. Returns True if the entire
            queue was sent"""
            assert fd == self.socket.fileno()
            while self.send_queue:
                self.sending = True
                try:
                    self.socket.send(self.send_queue.pop(0))
                except IOError, e:
                    if e.errno == errno.EWOULDBLOCK:
                        self.log.debug("got EWOULDBLOCK")
                        return False
                    elif e.errno in (errno.ECONNREFUSED, errno.ENOTCONN):
                        # If the worker has already shut down, and needs to be reaped,
                        # then we'll get back an ENOTCONN or an ECONNREFUSED from this
                        # send call. We should be reaping this process shortly hereafter
                        # anyway, so just ignore it
                        return False
                    else:
                        self.sending = False
                        raise
                except IndexError:
                    # We'll get the IndexError when we've sent the entire
                    # send_queue; if this ever gets multithreaded, we might
                    # get an IndexError instead of breaking out of the loop
                    # naturally, so gotta handle that!
                    break
                except Exception, e:
                    # Sometimes we seem to get errors that are not IOErrors?
                    if getattr(e, 'errno', None) not in (errno.ECONNREFUSED, errno.ENOTCONN):
                        raise
                    return False
            self.sending = False
            return True

        def sender(fd, *args, **kwargs):
            success = maybe_send_queue(fd, [])
            if success:
                self.io_loop.remove_handler(fd)

        # Try and send immediately
        success = maybe_send_queue(self.socket.fileno(), [])
        # if that fails, put it in the ioloop to send later
        if not success:
            self.io_loop.add_handler(self.socket.fileno(), sender, self.io_loop.WRITE)

    def send(self, msg):
        self.send_queue.append(msg)
        self._sendall()

class ZygoteIOLoop(tornado.ioloop.IOLoop):
    """Override IOLoop to log to our own logger instead of the root logger"""

    def __init__(self, *args, **kwargs):
        log_name = kwargs.pop('log_name', 'zygote.io_loop')
        self.log = logging.getLogger(log_name)
        super(ZygoteIOLoop, self).__init__(*args, **kwargs)

    def handle_callback_exception(self, callback):
        self.log.exception("Error in callback %s", callback)

    def add_handler(self, fd, handler, events):
        """Add a handler to the IOLoop, with exception handling"""

        # Get the list of available attributes from the handler.
        # A workaround for: http://bugs.python.org/issue3445
        available_attrs = (a for a in functools.WRAPPER_ASSIGNMENTS if hasattr(handler, a))

        @functools.wraps(handler, assigned=available_attrs)
        def wrapped_handler(*args, **kwargs):
            try:
                handler(*args, **kwargs)
            except Exception:
                self.handle_callback_exception(handler)

        return super(ZygoteIOLoop, self).add_handler(fd, wrapped_handler, events)



if hasattr(logging, 'NullHandler'):
    NullHandler = logging.NullHandler
else:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

# TODO: support logging to things other than stderr
#
# Why do we need this? Because we can not log in our signal handlers
# (http://docs.python.org/library/logging.html#thread-safety).
#
# This OTOH may cause log lines from master & worker processes merge
# into a single line. Happens very rarely and it's still better than
# master stuck in a deadlock.
class LocklessHandler(logging.StreamHandler):
    def createLock(self):
        self.lock = None

def get_logger(logger_name, debug=False):
    logger = logging.getLogger(logger_name)
    if os.isatty(sys.stderr.fileno()):
        formatter = logging.Formatter('[%(process)d] %(asctime)s :: %(levelname)-7s :: %(name)s - %(message)s')
        handler = LocklessHandler()
        handler.setFormatter(formatter)
        handler.setLevel(logging.DEBUG if debug else logging.INFO)
        logger.handlers = [handler]
        logger.propagate = False
    else:
        logger.handlers = [NullHandler()]
    return logger


def sanitize_headers(headers):
    """Sanitize sensitive request headers for logging"""
    results = dict(headers)
    # Redact instead of remove Authorization header so that those
    # using Basic Auth can debug if needed
    if results.get('Authorization'):
        results['Authorization'] = '***redacted***'
    return results



#!/usr/bin/env python

from setuptools import find_packages
from setuptools import setup


setup(
    name         = 'zygote',
    version      = '0.5.3',
    author       = 'Evan Klitzke',
    author_email = 'evan@eklitzke.org',
    description  = 'A tornado HTTP worker management tool',
    license      = 'Apache License 2.0',
    entry_points = {'console_scripts': 'zygote = zygote.main:main'},
    packages     = find_packages(exclude=['tests']),
    install_requires = ['setuptools', 'tornado'],
    include_package_data = True,
)



import logging
import time
import tornado.web

start_time = time.time()
log = logging.getLogger('example')
log.debug('started up')

class StatusHandler(tornado.web.RequestHandler):

    def get(self):
        self.content_type = 'text/plain'
        self.write('uptime: %1.3f\n' % (time.time() - start_time))

def initialize(*args, **kwargs):
    pass

def get_application(*args, **kwargs):
    log.debug('creating application for \'example\'')
    return tornado.web.Application([('/', StatusHandler)], debug=False)



from __future__ import with_statement

import datetime
import os
from pkg_resources import resource_filename
import socket
import time
import traceback

import tornado.httpserver
import tornado.web
import zygote.util

try:
    import simplejson as json
except ImportError:
    import json

class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if hasattr(obj, 'to_dict'):
            return obj.to_dict()
        elif type(obj) is datetime.datetime:
            return time.mktime(obj.timetuple()) + obj.microsecond / 1e6
        else:
            return super(JSONEncoder, self).default(obj)

class RequestHandler(tornado.web.RequestHandler):

    def get_error_html(self, status_code, **kwargs):
        if 500 <= status_code <= 599:
            self.set_header('Content-Type', 'text/plain')
            return traceback.format_exc()
        else:
            return super(RequestHandler, self).get_error_html(status_code, **kwargs)

class TemplateHandler(RequestHandler):

    def get(self):
        self.set_header('Content-Type', 'text/plain')
        self.set_header('Cache-Control', 'max-age=0')
        static_path = self.application.settings['static_path']
        with open(os.path.join(static_path, 'template.html')) as template:
            self.write(template.read())

class HTMLHandler(RequestHandler):

    def get(self):
        self.render('home.html')

class JSONHandler(RequestHandler):

    def get(self):

        self.zygote_master.zygote_collection.update_meminfo()
        env = self.zygote_master.zygote_collection.to_dict()
        env['hostname'] = socket.gethostname()
        env['interface'], env['port'] = self.application.settings['worker_sockname']
        env['pid'] = os.getpid()
        env['basepath'] = self.zygote_master.basepath
        env['time_created'] = self.zygote_master.time_created
        env.update(zygote.util.meminfo_fmt())

        self.set_header('Content-Type', 'application/json')
        self.write(json.dumps(env, cls=JSONEncoder, indent=2))

def get_httpserver(io_loop, port, zygote_master, zygote_base=None, ssl_options=None):
    if zygote_base is not None:
        static_path = os.path.realpath(os.path.join(zygote_base, 'zygote', 'resources', 'static'))
        template_path = os.path.realpath(os.path.join(zygote_base, 'zygote', 'resources', 'templates'))
    else:
        static_path = os.path.realpath(resource_filename('zygote.resources', 'static'))
        template_path = os.path.realpath(resource_filename('zygote.resources', 'templates'))

    # We need to ensure that we keep file handles open to the static path
    # and template path. If they go away (from some kind of clean up) while
    # the app is still running, we won't be able to serve the status page.
    # Bad!
    #
    # TODO: when implementing #24, these FDs will need to be cleaned up
    open_fds = []
    open_fds.append(os.open(static_path, os.O_DIRECTORY|os.O_RDONLY))
    open_fds.append(os.open(template_path, os.O_DIRECTORY|os.O_RDONLY))

    JSONHandler.zygote_master = zygote_master
    app = tornado.web.Application([('/', HTMLHandler),
                                   ('/json', JSONHandler),
                                   ('/template', TemplateHandler)],
                                  debug=False,
                                  static_path=static_path,
                                  template_path=template_path)
    app.settings['worker_sockname'] = zygote_master.sock.getsockname()
    http_server = tornado.httpserver.HTTPServer(app,
            io_loop=io_loop,
            no_keep_alive=True,
            ssl_options=ssl_options,
    )
    http_server.listen(port)
    return open_fds, http_server
